/*
 * Server configuration library implementation
 * File name: http_config.c
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: May 17, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "http_config.h"

static char *domains;
static int timeout;
static int maxthreads;
static char *logfilename;
static char *webdocsdir;
static char *errdocsdir;
static char *localdomainconffile;
static char *cgibindir;
static char *statspath;

// strings
#define LOGFILENAME     "log/access.log"
#define CONFFILE        "server_config/server.conf"
#define WEBDOCSDIR      "webdocs"
#define ERRDOCSDIR      "error"
#define CONFIGLOCALDOM  "server_config/local_domains"
#define CGIBINDIR       "cgi-bin"
#define STATSPATH       "/stats"

int set_option(char *line)
{
    char *p;

    if (line == NULL)
        return EXIT_FAILURE;
    else {
        if (*line == '#')
            return EXIT_SUCCESS;
        p = index(line,'=');
        if (p == NULL)
            return EXIT_FAILURE;
        if (strncasecmp(line,"timeout",p - line) == 0) {
            timeout = strtol(p + 1,NULL,10);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"maxthreads",p - line) == 0) {
            maxthreads = strtol(p + 1,NULL,10);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"logfilename",p - line) == 0) {
            logfilename = strndup(p + 1,strlen(p + 1) - 1);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"webdocsdir",p - line) == 0) {
            webdocsdir = strndup(p + 1,strlen(p + 1) - 1);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"errdocsdir",p - line) == 0) {
            errdocsdir = strndup(p + 1,strlen(p + 1) - 1);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"localdomainconffile",p - line) == 0) {
            localdomainconffile = strndup(p + 1,strlen(p + 1) - 1);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"cgibindir",p - line) == 0) {
            cgibindir = strndup(p + 1,strlen(p + 1) - 1);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"statspath",p - line) == 0) {
            statspath = strndup(p + 1,strlen(p + 1) - 1);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"port",p - line) == 0) {
            s_port = strtol(p + 1,NULL,10);
            return EXIT_SUCCESS;
        }
        if (strncasecmp(line,"uid",p - line) == 0) {
            uid = strtol(p + 1,NULL,10);
            return EXIT_SUCCESS;
        }
    }
    return EXIT_FAILURE;
}

int init_config(void)
{
    FILE *fp;
    ssize_t read;
    char *line = NULL;
    size_t len = 0;

    fp = fopen(CONFFILE,"r");
    if (fp == NULL)
        return EXIT_FAILURE;

    timeout = 60;            // default timeout for sockets
    maxthreads = 30;         // default max threads number
    logfilename = LOGFILENAME;
    webdocsdir = WEBDOCSDIR;
    errdocsdir = ERRDOCSDIR;
    localdomainconffile = CONFIGLOCALDOM;
    cgibindir = CGIBINDIR;
    statspath = STATSPATH;
    s_port = 80;             // default tcp port
    uid = getuid();          // set uid the same as the user who executed the server
    
    while ((read = getline(&line, &len, fp)) != -1)
        set_option(line);

    if (line)
        free(line);

    return EXIT_SUCCESS;
}

int get_maxthreads(void)
{
    return maxthreads;
}

u_char *get_webdocsdir(void)
{
    return webdocsdir;
}

u_char *get_errdocsdir(void)
{
    return errdocsdir;
}

u_char *get_cgidir(void)
{
    return cgibindir;
}

u_char *get_logfile(void)
{
    return logfilename;
}

int get_timeout(void)
{
    return timeout;
}

char *get_statspath(void)
{
    return statspath;
}

ssize_t set_localdomain(void)
{
    int fd;
    ssize_t n;
    struct stat *stats;
    u_char *ptr;
    u_char *str_ip;

    if ((str_ip = inet_ntoa(s_addr)) == NULL) {
        fprintf(stderr,"set_localdomain(): %s\n",strerror(errno));
        _exit(-1);
    }

    if ((stats = malloc(sizeof(struct stat))) == NULL) {
        fprintf(stderr,"set_localdomain(): %s\n",strerror(errno));
        _exit(-1);
    }
    bzero(stats,sizeof(struct stat));
    stat(localdomainconffile,stats);
    if (!S_ISREG(stats->st_mode)) {
        free(stats);
        fprintf(stderr,"set_localdomain(): %s %s\n",strerror(errno),localdomainconffile);
        _exit(-1);
    }

    if ((domains = malloc((stats->st_size + 1 + strlen(str_ip))*sizeof(u_char))) == NULL) {
        free(stats);
        fprintf(stderr,"set_localdomain(): %s\n",strerror(errno));
        _exit(-1);
    }
    bzero(domains,stats->st_size + 1);

    fd = open(localdomainconffile,O_RDONLY);
    if (fd == -1) {
        free(stats);
        free(domains);
        fprintf(stderr,"set_localdomain(): %s %s\n",strerror(errno),localdomainconffile);
        _exit(-1);
    }

    if (stats->st_size == 0) {
        free(stats);
        free(domains);
        fprintf(stderr,"set_localdomain(): no domains in %s file\n",localdomainconffile);
        _exit(-1);
    }

    while ((n = read(fd,domains,stats->st_size)) != stats->st_size)
        if ((n == -1) && (errno != EINTR)) {
            free(stats);
            free(domains);
            fprintf(stderr,"set_localdomain(): %s\n",strerror(errno));
            _exit(-1);
        }
    strncat(domains,str_ip,strlen(str_ip));

    ptr = index(domains,'\n');
    if (ptr == NULL)
        default_domain = domains;
    else
        if ((default_domain = strndup(domains,strlen(domains) - strlen(ptr))) == NULL) {
            free(stats);
            free(domains);
            fprintf(stderr,"set_localdomain(): %s\n",strerror(errno));
            _exit(-1);
        }

    free(stats);
    close(fd);

    return n;
}

int is_localdomain(const u_char *str)
{
    u_char *ptrlf, *ptrline;

    assert(str != NULL);
    ptrline = domains;

    if (strncasecmp(str,inet_ntoa(s_addr),strlen(str)) == 0)
        return 1;

    while (1) {
        ptrlf = index(ptrline,'\n');
        if (ptrlf == NULL) {
            if (*ptrline != '\0')
                return (strncasecmp(str,ptrline,strlen(ptrline)) == 0) ? 1 : 0;
            else 
                return 0;
        }
        if (strncasecmp(str,ptrline,ptrlf - ptrline) == 0)
            return 1;
        ptrline = ptrlf + 1;
    }
}

int is_stats(const u_char *str)
{
    if (str == NULL)
        return 0;

    return (strncasecmp(str,statspath,strlen(statspath)) == 0) ? 1 : 0;
}

