/*
 * Example of a simple HTTP server.
 * File name: http_server.c
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: June 22, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "defs.h"
#include "message_analysis.h"
#include "http_10.h"
#include "http_11.h"
#include "proxy.h"
#include "http_config.h"
#include "utils.h"
#include "statistics.h"

static int threads = 0;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

void usage(char **argv)
{
    fprintf(stderr,"%s\n"
            "usage: %s host [port] [uid]\n"
            "\thost\thost name or address to bind to\n"
            "\t[port]\ttcp port where to listen\n"
            "\t[uid]\toptional user id to run as unprivileged\n",SERVERNAME,argv[0]);
    _exit(EXIT_FAILURE);
}

void *attend(void *arg)
{
    u_char request[MAXBUFSIZE];
    int conn_fd, found, exit, n, pos;
    petition_t *petition;
    struct timeval *t;
    //struct sockaddr_in *sa;

    exit = 1;        // this is the first petition
    conn_fd = (int) arg;
    pthread_detach(pthread_self());	// mark as not joinable

    //n = sizeof(struct sockaddr_in);
    //sa = malloc(n);
    // this is a dirty hack, getpeername expects struct sockaddr, not struct sockaddr_in
    //getpeername(conn_fd,(struct sockaddr *) sa,&n);
    //printf("Connection accepted from %s\n",inet_ntoa(sa->sin_addr));
    //free(sa);

    // set socket timeout
    if ((t = malloc(sizeof(struct timeval))) == NULL)
        pthread_exit(NULL);
    bzero(t,sizeof(struct timeval));
    t->tv_sec = get_timeout();
    t->tv_usec = 0;
    setsockopt(conn_fd,SOL_SOCKET,SO_RCVTIMEO,t,sizeof(struct timeval));
    setsockopt(conn_fd,SOL_SOCKET,SO_SNDTIMEO,t,sizeof(struct timeval));
    free(t);

    while (exit <= 5) {		// allow max 5 petitions on a persistent conn
        bzero(request,MAXBUFSIZE);

        // read petition
        if (readn(conn_fd, request, MAXBUFSIZE) <= 0)
            break;

        //printf(">--------------------------------------------------------------------------------\n");
        //printf("Petition: %s",request);
        if ((petition = init_petition()) == NULL) {
            fprintf(stderr,"attend(): %s\n",strerror(errno));
            break;
        }

        // attend petition
        found = 0;
        if (analyze_petition(request,petition) == EBADSYNTAX) {
            write_log(conn_fd,petition,BADREQUEST,"default");
            n = http_10_errmsg(conn_fd,BADREQUEST,NULL,IFCLOSE,petition->request.method);
            free_petition(&petition);
            break;
        } else {
            // read body if present
            pos = search_header(petition,PETITION,CONTENTLEN);
            if ((pos != -1) && (pos != petition->headerno)) {
                if ((petition->body = malloc((strtol(petition->header[pos].value,NULL,10) +1)*sizeof(char))) == NULL) {
                    free_petition(&petition);
                    pthread_exit(NULL);
                }
                bzero(petition->body,strtol(petition->header[pos].value,NULL,10) + 1);
                if ((n = readn(conn_fd, petition->body, strtol(petition->header[pos].value,NULL,10))) <= 0)
                    break;

                *(petition->body + (n+1)*sizeof(u_char)) = '\0';
            }
            //printf("Body: %s\n\n",petition->body);

            // HTTP/1.0 petition
            if (petition->request.version == HTTP10) {
                // should keep connection alive
                if (H_KEEPALIVE(petition->flags)) {
                    // proxy petition
                    if (is_proxy(petition)) {
                        /* we should keep connection open, so close only
                           if failed to attend petition */
                        if (attend_remote_petition(conn_fd,petition,0)) {
                            free_petition(&petition);
                            break;
                        }
                    // local petition
                    } else {
                        /* we should keep connection open, so close only
                           if failed to attend petition */
                        if (attend_10_petition(conn_fd,petition,0)) {
                            free_petition(&petition);
                            break;
                        }
                    }
                    exit++;
                // should close connection
                } else {
                    // proxy petition, close always
                    if (is_proxy(petition)) {
                        attend_remote_petition(conn_fd,petition,IFCLOSE);
                        free_petition(&petition);
                        break;
                    }
                    // local petition, close always
                    attend_10_petition(conn_fd,petition,IFCLOSE);
                    free_petition(&petition);
                    break;
                }
            }
            // HTTP/1.1 petition
            if (petition->request.version == HTTP11) {
                // should close connection
                if (H_CLOSE(petition->flags)) {
                    // if precondition failed, close
                    if (!H_HOST(petition->flags)) {
                        write_log(conn_fd,petition,BADREQUEST,"default");
                        n = http_11_errmsg(conn_fd,BADREQUEST,NULL,IFCLOSE,petition->request.method);
                        free_petition(&petition);
                        break;
                    }
                    // proxy petition, close always
                    if (is_proxy(petition)) {
                        attend_remote_petition(conn_fd,petition,IFCLOSE);
                        free_petition(&petition);
                        break;
                    }
                    // local petition, close always
                    attend_11_petition(conn_fd,petition,IFCLOSE);
                    free_petition(&petition);
                    break;
                // should not close connection
                } else {
                    // precondition failed
                    if (!H_HOST(petition->flags)) {
                        write_log(conn_fd,petition,BADREQUEST,"default");
                        n = http_11_errmsg(conn_fd,BADREQUEST,NULL,0,petition->request.method);
                    } else {
                        // proxy petition, close if fail
                        if (is_proxy(petition)) {
                            if (attend_remote_petition(conn_fd,petition,0)) {
                                free_petition(&petition);
                                break;
                            }
                        // local petition, close if fail
                        } else {
                            if (attend_11_petition(conn_fd,petition,0)) {
                                free_petition(&petition);
                                break;
                            }
                        }
                    }
                    exit++;
                }
            }
            // UNKNOWN PROTOCOL VERSION
            if (petition->request.version == UNKNOWN) {
                write_log(conn_fd,petition,BADREQUEST,"default");
                n = http_10_errmsg(conn_fd,BADREQUEST,NULL,IFCLOSE,petition->request.method);
                free_petition(&petition);
                break;
            }
        }
        free_petition(&petition);
    }
    // we are going to exit, one less thread
    pthread_mutex_lock(&mutex);
    threads--;
    pthread_mutex_unlock(&mutex);

    close(conn_fd);
    pthread_exit(NULL);
}

int main (int argc, char *argv[])
{
    int listen_fd, conn_fd, reuse;
    struct sockaddr_in client, server;	// client and server end points
    int cli_len;	// client sockaddr struct size
    struct hostent *he;
    pthread_t tid;
    
    // check if required arguments are present
    if (argc < 2)
        usage(argv);

    // init server config
    if (init_config() == EXIT_FAILURE)
        fprintf(stderr,"init_config(): %s\n",strerror(errno));

    // get ip from name
    if ((he = gethostbyname(argv[1])) == NULL) {
        fprintf(stderr,"nslookup failed on %s\n",argv[1]);
        exit(-1);
    }
    s_addr.s_addr = *((unsigned long *) he->h_addr_list[0]);

    // get server local domain
    set_localdomain();

    if (argc >= 3)
        s_port = strtol(argv[2], NULL, 10);

    // create socket
    listen_fd = socket(AF_INET, SOCK_STREAM, 0);

    // set socket options
    reuse = 1;
    setsockopt(listen_fd,SOL_SOCKET,SO_REUSEADDR,&reuse,sizeof(reuse));

    // make endpoint
    bzero(&server, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = htonl(INADDR_ANY);
    server.sin_port = htons(s_port);

    // bind server to proposed addr
    if (bind(listen_fd, (SA *) &server, sizeof(server)) == -1) {
        printf("bind error: %s\n", strerror(errno));
        exit(-1);
    }

    // listen into binded end point
    if (listen(listen_fd, SOMAXCONN) == -1) {
        printf("listen error: %s\n", strerror(errno));
        exit(-1);
    }

    // drop privileges if there is a user defined to run the server
    if (argc == 4)
        uid = strtol(argv[3], NULL, 10);
    if (setuid(uid) == -1)
        fprintf(stderr,"setuid(%d): %s\n",uid,strerror(errno));
    if (seteuid(uid) == -1)
        fprintf(stderr,"seteuid(%d): %s\n",uid,strerror(errno));


    init_stats();
    init_logs();
    
    printf("%s started at %s (%s:%ld) with UID %d\n",
            SERVERNAME,
            getenv("HOSTNAME"),
            inet_ntoa(s_addr),
            s_port,
            getuid());
    printf("Waiting for remote connection\n");
    
    for (;;) {
        cli_len = sizeof(&client);
        pthread_mutex_lock(&mutex);
        if (threads < get_maxthreads()) {
            pthread_mutex_unlock(&mutex);
            if ((conn_fd = accept(listen_fd, (SA *) &client, &cli_len)) < 0) {
                if (errno == EINTR)
                    continue;	// try again
                else {
                    fprintf(stderr,"Can't accept incoming connection%s\n", strerror(errno));
                    exit(-1);
                }
            }
            pthread_mutex_lock(&mutex);
            threads++;
            pthread_mutex_unlock(&mutex);
            pthread_create(&tid, NULL, attend, (void *) conn_fd);	// attend in a new thread
        } else {
            pthread_mutex_unlock(&mutex);
            sched_yield();	// give processor to threads, wait until someone finishes
        }
    }

    close(listen_fd);
}
