/*
 * HTTP message analysis library implementation
 * File name: message_analysis.c
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: May 18, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "message_analysis.h"
#include "http_config.h"
#include "http_10.h"
#include "http_11.h"

/*
 * PETITION ANALYSIS FUNCS
 */

int analyze_petition(const u_char *request, petition_t petition)
{
    int requestlen, headerrslt, pos;
    u_char *ptrcrlf, *ptrcrlfaux, *requestline, *headerline, *ptr;
    request_t req = NULL;
    header_t hdr = NULL;

    assert(request != NULL);
    if ((requestlen = strlen(request)) == 0)
        return EBADSYNTAX;

    assert(petition != NULL);

    /* search for request line and parse it */
    ptrcrlf = index(request,'\r');
    if (ptrcrlf == NULL)
        return EBADSYNTAX;

    if (strncmp(ptrcrlf,"\r\n",2) == 0) {
        if ((requestline = strndup(request,(requestlen - strlen(ptrcrlf) + 2))) == NULL)
            return ENOTENMEM;

        if ((req = malloc(sizeof(struct request))) == NULL) {
            free(requestline);
            return ENOTENMEM;
        }
        bzero(req,sizeof(struct request));
        if (break_request_line(requestline,req) == EBADSYNTAX) {
            free(requestline);
            free(req);
            return EBADSYNTAX;
        }
        free(requestline);
    } else
        return EBADSYNTAX;

    /* fill in petition structure */
    petition->request.method = req->method;
    petition->request.version = req->version;
    petition->request.uri.protocol = req->uri.protocol;
    petition->request.uri.host = req->uri.host;
    petition->request.uri.path = req->uri.path;
    petition->request.uri.port = req->uri.port;
    petition->request.uri.port_present = req->uri.port_present;
    petition->headerno = 0;
    free(req);

    /* search for headers */
    ptrcrlfaux = index((char *) ptrcrlf +2,'\r');
    while (ptrcrlfaux != NULL) {
        /* extract header line */
        if (strlen(ptrcrlf +2) > strlen(ptrcrlfaux)) {
            if ((headerline = strndup((char *) ptrcrlf +2,
                                 strlen(ptrcrlf +2) - strlen(ptrcrlfaux))) == NULL) {
                free(req->uri.protocol);
                free(req->uri.path);
                free(req->uri.host);
                return ENOTENMEM;
            }
            /* parse header line */
            if ((hdr = malloc(sizeof(struct header))) == NULL) {
                free(headerline);
                free(req->uri.protocol);
                free(req->uri.path);
                free(req->uri.host);
                return ENOTENMEM;
            }
            bzero(hdr,sizeof(struct header));
            headerrslt = break_header(headerline,hdr,&petition->flags);
            if (headerrslt == EBADSYNTAX) {
                free(hdr->value);
                free(hdr);
                return EBADSYNTAX;
            }

            if (hdr->name == UNKNOWN)
                pos = add_header(petition,PETITION,hdr->name,(const u_char *) headerline);
            else
                pos = add_header(petition,PETITION,hdr->name,(const u_char *) hdr->value);

            free(headerline);
            free(hdr->value);
            free(hdr);

            /* search for new header */
            if (strncmp(ptrcrlfaux,"\r\n",2) != 0)
                return EBADSYNTAX;

            ptr = index((u_char *) ptrcrlfaux +2,'\r');
            ptrcrlf = ptrcrlfaux;
            ptrcrlfaux = ptr;
        } else
            /* length is equal, \r\n\r\n found? */
            ptrcrlfaux = NULL;
    }
    /* finished examining headers, body? */
    if (strncmp(ptrcrlf,"\r\n\r\n",4) != 0)    /* we know ptrcrlf is at least \r\n\n\0 */
        return EBADSYNTAX;
    
    ptrcrlf += 4;
    if (*ptrcrlf == '\0')
        petition->body = NULL;
    else
        if ((petition->body = strndup(ptrcrlf,strlen(ptrcrlf))) == NULL) {
            free(req->uri.protocol);
            free(req->uri.path);
            free(req->uri.host);
            return ENOTENMEM;
        }
    return 0;
}


int break_request_line(const u_char *line, request_t request)
{
    int linelen, version, found;
    u_char *ptrb, *ptrmethod, *ptruri, *ptrversion, *ptruritmp;

    assert(line != NULL);

    linelen = strlen(line);

    if (linelen == 0)
        return EBADSYNTAX;
    if (request == NULL)
        return ENALLOC;

    ptrb = index(line,' ');    /* search for a white */
    if (ptrb == NULL)
        return EBADSYNTAX;

    /* take method apart */
    if ((ptrmethod = strndup(line,strlen(line) - strlen(ptrb))) == NULL)
        return ENOTENMEM;
    
    /* trim white spaces before uri begins */
    while (*ptrb == ' ')
        if (strlen(ptrb) > 0)
            ptrb++;

    if (*(ptruritmp = ptrb) == '\0')    /* if empty, return */
        return EBADSYNTAX;

    ptrb = index(ptruritmp,' ');    /* search for a new white space */
    if (ptrb == NULL)
        return EBADSYNTAX;

    if ((ptruri = strndup(ptruritmp,strlen(ptruritmp) - strlen(ptrb))) == NULL) {
        free(ptrmethod);
        return ENOTENMEM;
    }

    /* decompound uri into fields */
    if (break_uri(ptruri,&request->uri) != 0)
        return EBADSYNTAX;

    /* trim whites */
    while (*ptrb == ' ')
        if (strlen(ptrb) > 0)
            ptrb++;

    /* check version */
    if (*(ptrversion = ptrb) == '\0')    /* empty? */
        return EBADSYNTAX;

    if (strncasecmp(ptrversion,"HTTP/1.0\r\n",strlen(ptrversion)) == 0)
        version = HTTP10;
    else if (strncasecmp(ptrversion,"HTTP/1.1\r\n",strlen(ptrversion)) == 0)
        version = HTTP11;
    else
        version = UNKNOWN;

    found = 0;
    /* finally, check if method is ok */
    if (!found && strncasecmp(ptrmethod,"OPTIONS",strlen(ptrmethod)) == 0) {
        request->method = OPTIONS;
        found = 1;
    }
    if (!found && strncasecmp(ptrmethod,"GET",strlen(ptrmethod)) == 0) {
        request->method = GET;
        found = 1;
    }
    if (!found && strncasecmp(ptrmethod,"HEAD",strlen(ptrmethod)) == 0) {
        request->method = HEAD;
        found = 1;
    }
    if (!found && strncasecmp(ptrmethod,"POST",strlen(ptrmethod)) == 0) {
        request->method = POST;
        found = 1;
    }
    if (!found && strncasecmp(ptrmethod,"PUT",strlen(ptrmethod)) == 0) {
        request->method = PUT;
        found = 1;
    }
    if (!found && strncasecmp(ptrmethod,"DELETE",strlen(ptrmethod)) == 0) {
        request->method = DELETE;
        found = 1;
    }
    if (!found && strncasecmp(ptrmethod,"TRACE",strlen(ptrmethod)) == 0) {
        request->method = TRACE;
        found = 1;
    }
    if (!found && strncasecmp(ptrmethod,"CONNECT",strlen(ptrmethod)) == 0) {
        request->method = CONNECT;
        found = 1;
    }

    if (!found)
        request->method = UNKNOWN;

    request->version = version;
    free(ptrmethod);
    free(ptruri);
    return 0;
}

int check_host(const petition_t petition, int port)
{
    u_char *ptrdots;
    int pos;

    pos = search_header(petition,PETITION,HOST);
    // we have at least a host in uri
    if (petition->request.uri.host != NULL) {
        if (port)
            if (petition->request.uri.port != port)
                return 0;	// bad port!
        // we also have a host header
        if ((pos != petition->headerno) && (pos != -1)) {
            if ((strncasecmp(petition->request.uri.host,
                            petition->header[pos].value,
                            strlen(petition->request.uri.host))) != 0)
                return 0;    // hosts do not match, false
            /* check if we have a port specified in host header */
            if ((ptrdots = index(petition->header[pos].value,':')) != NULL) {
                if (ptrdots != (petition->header[pos].value +
                                strlen(petition->request.uri.host)*sizeof(u_char)))
                     return 0;/* there is a port in header, but host is longer than that
                                 specified in uri */
                if (strtol(ptrdots + 1*sizeof(u_char),NULL,10) != petition->request.uri.port)
                    return 0;    // port in header is not equal to uri port
            } else
                if (strlen(petition->header[pos].value) != strlen(petition->request.uri.host))
                    return 0;    // header length is not equal to uri length
        }
    // no host in uri
    } else
        // we have host header
        if ((pos != petition->headerno) && (pos != -1))
            // port?
            if ((ptrdots = index(petition->header[pos].value,':')) != NULL)
                if (strtol(ptrdots + 1*sizeof(u_char),NULL,10) != s_port)
                    return 0;    // this is not proxy, and we are not listening there!

    return 1;	// all is ok here
}

u_char *get_host(const petition_t petition)
{
    int pos, len = 0;
    u_char *ptrdots;
    char *ip;

    assert(petition != NULL);

    ip = inet_ntoa(s_addr);

    if (petition->request.uri.host != NULL)
        if (strncmp(petition->request.uri.host,ip,strlen(petition->request.uri.host)) != 0)
            return strndup(petition->request.uri.host,strlen(petition->request.uri.host));
    pos = search_header(petition,PETITION,HOST);
    if ((pos != -1) && (pos != petition->headerno)) {
        if ((ptrdots = index(petition->header[pos].value,':')) != NULL)
            len = strlen(ptrdots);
        if (strncmp(petition->header[pos].value,ip,strlen(petition->header[pos].value) - len) != 0)
            return strndup(petition->header[pos].value,strlen(petition->header[pos].value) - len);
    }

    return strndup(default_domain,strlen(default_domain));
}

/*
 * PETITION STRUCTURE MANAGEMENT
 */

petition_t init_petition(void)
{
    petition_t petition;
    int i;

    if ((petition = malloc(sizeof(struct petition))) == NULL)
        return NULL;

    bzero(petition,sizeof(struct petition));
    if (petition != NULL) {
        petition->request.method = -1;
        petition->request.version = -1;
        petition->request.uri.protocol = NULL;
        petition->request.uri.host = NULL;
        petition->request.uri.port = -1;
        petition->request.uri.port_present = 0;
        petition->request.uri.path = NULL;
        petition->flags = 0;
        petition->headerno = 0;	/* zero headers by default */
        for (i = 0; i < MAXHEADERS; i++) {
            petition->header[i].name = EMPTY;
            petition->header[i].value = NULL;
        }
        petition->body = NULL;
    }

    return petition;
}

int free_petition(petition_t *petition)
{
    int i;
    petition_t p;
    
    assert(petition != NULL);
    p = *petition;
    assert(*petition != NULL);

    if (p->request.uri.protocol != NULL)
        free(p->request.uri.protocol);
    if (p->request.uri.host != NULL)
        free(p->request.uri.host);
    if (p->request.uri.path != NULL)
        free(p->request.uri.path);
    if (p->body != NULL)
        free(p->body);
    if (p->headerno > 0)
        for (i=0; i < p->headerno; i++)
            if (p->header[i].value != NULL)
                free(p->header[i].value);
    free(*petition);

    return 0;
}

u_char *sprintp(petition_t petition)
{
    u_char *buffer, uri[MAXBUFSIZE];
    int n, i;

    assert(petition != NULL);
    if ((buffer = malloc(MAXBUFSIZE*sizeof(u_char))) == NULL)
        return NULL;

    bzero(buffer,MAXBUFSIZE);

/*    snprintf(uri,MAXBUFSIZE,"http://%s:%d%s",
             petition->request.uri.host,
             petition->request.uri.port,
             petition->request.uri.path);*/
    /* 
     * Some servers do not accept absolute uri's.
     * Comment the following and uncomment the previous lines
     * to use absolute uri's on forwarded queries.
     */
    snprintf(uri,MAXBUFSIZE,"%s",petition->request.uri.path);

    // print request line
    if (petition->request.version == HTTP11) {
        n = http_11_request(buffer,petition->request.method,uri,MAXBUFSIZE);
        for (i = 0; i < petition->headerno; i++)
            if ((petition->header[i].name != EMPTY) &&
                (petition->header[i].value != NULL))
                http_11_header(buffer,
                               petition->header[i].name,
                               MAXBUFSIZE - strlen(buffer),
                               petition->header[i].value);
    }
    if (petition->request.version == HTTP10) {
        n = http_10_request(buffer,petition->request.method,uri,MAXBUFSIZE);
        for (i = 0; i < petition->headerno; i++)
            if ((petition->header[i].name != EMPTY) &&
                (petition->header[i].value != NULL))
                http_11_header(buffer,
                               petition->header[i].name,
                               MAXBUFSIZE - strlen(buffer),
                               petition->header[i].value);
    }

    n = crlf(buffer,MAXBUFSIZE - strlen(buffer));

    return buffer;
}

/*
 * HEADERS MANAGEMENT FUNCTIONS
 */

int break_header(const u_char *line, header_t header, int *flag)
{
    int linelen, namelen, valuelen, found, n;
    u_char *ptrdots, *name, *value, *ptraux;
    
    assert(line != NULL);
    
    linelen = strlen(line);
    assert(linelen != 0);
    assert(header != NULL);

    ptrdots = index(line,':');	/* look for value separator */
    if (ptrdots == NULL)
        return EBADSYNTAX;

    /* separate field name and value into different strings */
    if ((name = strndup(line,strlen(line) - strlen(ptrdots))) == NULL)
        return ENOTENMEM;

    /* set ptraux and value, ptraux will be const, value will be used to trim
       whites */
    if ((ptraux = strndup((char *) ptrdots +1,strlen(ptrdots) -1)) == NULL) {
        free(name);
        return ENOTENMEM;
    }
    value = ptraux;

    /* trim value for possible begining blanks */
    while (*value == ' ')
        if (strlen(value) > 0)
            value++;
    /* if empty value (\0), then make value null */
    if (*value == '\0')
        value = NULL;
     
     namelen = strlen(name);
     valuelen = 0;
     if (value != NULL)
         valuelen = strlen(value);

    found = 0;
    /* check if name is a valid header name. header->value is not set
       until we are sure it's a valid header line. Because of efficiency,
       return inmediately we have found header is OK. */
    if (!found && strncasecmp(name,"Cache-Control",namelen) == 0) {
        n = CACHECONTROL;
        found = 1;
    }
    if (!found && strncasecmp(name,"Connection",namelen) == 0){
        n = CONNECTION;
        if (strncasecmp(value,"Keep-Alive",valuelen) == 0)
            *flag |= IFKEEPALIVE;
        if (strncasecmp(value,"Close",valuelen) == 0)
            *flag |= IFCLOSE;
        found = 1;
    }
    if (!found && strncasecmp(name,"Date",namelen) == 0) {
        n = DATE;
        found = 1;
    }
    if (!found && strncasecmp(name,"Pragma",namelen) == 0) {
        n = PRAGMA;
        found = 1;
    }
    if (!found && strncasecmp(name,"Trailer",namelen) == 0) {
        n = TRAILER;
        found = 1;
    }
    if (!found && strncasecmp(name,"Transfer-Encoding",namelen) == 0) {
        n = TRANSFERENC;
        found = 1;
    }
    if (!found && strncasecmp(name,"Upgrade",namelen) == 0) {
        n = UPGRADE;
        found = 1;  
    }
    if (!found && strncasecmp(name,"Via",namelen) == 0) {
        n = VIA;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Warning",namelen) == 0) {
        n = WARNING;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Accept",namelen) == 0) {
        n = ACCEPT;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Accept-Charset",namelen) == 0) {
        n = ACCEPTCHARSET;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Accept-Encoding",namelen) == 0) {
        n = ACCEPTENC;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Accept-Language",namelen) == 0) {
        n = ACCEPTLANG;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Authorization",namelen) == 0) {
        *flag |= IFAUTH;
        n = AUTHORIZATION;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Expect",namelen) == 0) {
        n = EXPECT;
        found = 1;   
    }
    if (!found && strncasecmp(name,"From",namelen) == 0) {
        n = FROM;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Host",namelen) == 0) {
        *flag |= IFHOST;
        n = HOST;
        found = 1;   
    }
    if (!found && strncasecmp(name,"If-Match",namelen) == 0) {
        n = IFMATCH;
        found = 1;   
    }
    if (!found && strncasecmp(name,"If-Modified-Since",namelen) == 0) {
        n = IFMODIFSINCE;
        found = 1;   
    }
    if (!found && strncasecmp(name,"If-None-Match",namelen) == 0) {
        n = IFNONEMATCH;
        found = 1;   
    }
    if (!found && strncasecmp(name,"If-Range",namelen) == 0) {
        n = IFRANGE;
        found = 1;   
    }
    if (!found && strncasecmp(name,"If-Unmodified-Since",namelen) == 0) {
        n = IFUNMODSINCE;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Max-Forwards",namelen) == 0) {
        n = MAXFORWARDS;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Proxy-Authorization",namelen) == 0) {
        n = PROXYAUTHOR;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Range",namelen) == 0) {
        n = RANGE;
        *flag |= IFRANGEH;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Referer",namelen) == 0) {
        n = REFERER;
        found = 1;   
    }
    if (!found && strncasecmp(name,"TE",namelen) == 0) {
        n = TE;
        found = 1;   
    }
    if (!found && strncasecmp(name,"User-Agent",namelen) == 0) {
        n = USERAGENT;
        found = 1;    
    }
    if (!found && strncasecmp(name,"Allow",namelen) == 0) {
        n = ALLOW;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Content-Encoding",namelen) == 0) {
        n = CONTENTENC;
        found = 1;    
    }
    if (!found && strncasecmp(name,"Content-Language",namelen) == 0) {
        n = CONTENTLANG;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Content-Length",namelen) == 0) {
        n = CONTENTLEN;
        found = 1;
        *flag |= IFCONTLEN;
    }
    if (!found && strncasecmp(name,"Content-Location",namelen) == 0) {
        n = CONTENTLOC;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Content-MD5",namelen) == 0) {
        n = CONTENTMD5;
        found = 1;
    }
    if (!found && strncasecmp(name,"Content-Range",namelen) == 0) {
        n = CONTENTRANGE;
        found = 1;   
    }
    if (!found && strncasecmp(name,"Content-Type",namelen) == 0) {
        n = CONTENTTYPE;
        found = 1;
    }
    if (!found && strncasecmp(name,"Expires",namelen) == 0) {
        n = EXPIRES;
        found = 1;
    }
    if (!found && strncasecmp(name,"Last-Modified",namelen) == 0) {
        n = LASTMODIFIED;
        found = 1;
    }
    if (!found && strncasecmp(name,"Location",namelen) == 0) {
        n = LOCATION;
        found = 1;
    }
    if (!found && strncasecmp(name,"Server",namelen) == 0) {
        n = SERVER;
        found = 1;
    }
    if (!found && strncasecmp(name,"WWW-Authenticate",namelen) == 0) {
        n = WWWAUTH;
        found = 1;
    }
    if (!found && strncasecmp(name,"Accept-Ranges",namelen) == 0) {
        n = ACCEPTRANGES;
        found = 1;
    }
    if (!found && strncasecmp(name,"Age",namelen) == 0) {
        n = AGE;
        found = 1;
    }
    if (!found && strncasecmp(name,"ETag",namelen) == 0) {
        n = ETAG;
        found = 1;
    }
    if (!found && strncasecmp(name,"Proxy-Authenticate",namelen) == 0) {
        n = PROXYAUTHEN;
        found = 1;
    }
    if (!found && strncasecmp(name,"Retry-After",namelen) == 0) {
        n = RETRYAFTER;
        found = 1;
    }
    if (!found && strncasecmp(name,"Vary",namelen) == 0) {
        n = VARY;
        found = 1;
    }
    if (!found && strncasecmp(name,"Proxy-Connection",namelen) == 0) {
        n = PROXYCONN;
        if (strncasecmp(value,"Keep-Alive",valuelen) == 0)
            *flag |= IFKEEPALIVE;
        if (strncasecmp(value,"Close",valuelen) == 0)
            *flag |= IFCLOSE;
        found = 1;
    }
    if (!found && strncasecmp(name,"Keep-Alive",namelen) == 0) {
        n = KEEPALIVE;
        found = 1;
    }

    if (!found)
        n = UNKNOWN;
    header->name = n;

    if (valuelen)
        if ((header->value = strndup(value,valuelen)) == NULL) {
            free(name);
            free(ptraux);
            return ENOTENMEM;
        }

    free(name);
    free(ptraux);

    return 0;
}
 
int add_header(void *message, const int type, const int name, const u_char *value)
{
    petition_t petition = NULL;
    reply_t reply = NULL;

    if (message == NULL)
        return -1;

    switch (type) {
    case PETITION:
        petition = (petition_t) message;
        if (petition->headerno < MAXHEADERS) {
            petition->header[petition->headerno].name = name;
            petition->header[petition->headerno].value = NULL;
            if (value != NULL)
                if ((petition->header[petition->headerno].value = strndup(value,strlen(value))) == NULL)
                    return ENOTENMEM;
            petition->headerno++;
            break;
        } else
            return -1;
    case REPLY:
        reply = (reply_t) message;
        if (reply->headerno < MAXHEADERS) {
            reply->header[reply->headerno].name = name;
            reply->header[reply->headerno].value = NULL;
            if (value != NULL)
                if ((reply->header[reply->headerno].value = strndup(value,strlen(value))) == NULL)
                    return ENOTENMEM;
            reply->headerno++;
            break;
        } else
            return -1;
    default:
        return -1;
    }
    return 0;
}

int search_header(void *message, const int type, const int name)
{
    petition_t petition = NULL;
    reply_t reply = NULL;
    int i, position = 0;

    if (message == NULL)
        return -1;

    switch (type) {
    case PETITION:
        petition = (petition_t) message;
        for (i = 0; i < petition->headerno; i++) {
            if (petition->header[i].name == name)
                break;
            position++;
        }
        break;
    case REPLY:
        reply = (reply_t) message;
        for (i = 0; i < reply->headerno; i++) {
            if (reply->header[i].name == name)
                break;
            position++;
        }
        break;
    default:
        return -1;
    }

    return position;
}

int remove_header(void *message, const int type, const int name)
{
    petition_t petition = NULL;
    reply_t reply = NULL;
    int i;

    switch (type) {
    case PETITION:
        petition = (petition_t) message;
        for (i = 0; i < petition->headerno; i++)
            if (petition->header[i].name == name) {
                if (petition->header[i].value != NULL)
                    free(petition->header[i].value);
                petition->header[i].name = petition->header[petition->headerno - 1].name;
                petition->header[i].value = petition->header[petition->headerno - 1].value;
                petition->header[petition->headerno - 1].name = EMPTY;
                petition->header[petition->headerno - 1].value = NULL;
                petition->headerno--;
                break;
            }
        break;
    case REPLY:
        reply = (reply_t) message;
        for (i = 0; i < reply->headerno; i++)
            if (reply->header[i].name == name) {
                if (reply->header[i].value != NULL)
                    free(reply->header[i].value);
                reply->header[i].name = reply->header[reply->headerno - 1].name;
                reply->header[i].value = reply->header[reply->headerno - 1].value;
                reply->header[reply->headerno - 1].name = EMPTY;
                reply->header[reply->headerno - 1].value = NULL;
                reply->headerno--;
                break;
            }
        break;
    default:
        return -1;
    }
    return 0;    
}

/*
 * REPLY MANAGEMENT FUNCS
 */

int analyze_reply(const u_char *response, reply_t reply)
{
    int responselen, headerrslt, pos;
    u_char *ptrcrlf, *ptrcrlfaux, *responseline, *headerline, *ptr;
    header_t hdr = NULL;
    status_t status = NULL;

    assert(response != NULL);
    if ((responselen = strlen(response)) == 0)
        return EBADSYNTAX;

    assert(reply != NULL);

    /* search for request line and parse it */
    ptrcrlf = strstr(response,"\r\n");
    if (ptrcrlf == NULL)
        return EBADSYNTAX;

    if ((responseline = strndup(response,(responselen - strlen(ptrcrlf)))) == NULL)
        return ENOTENMEM;

    if ((status = malloc(sizeof(struct status))) == NULL) {
        free(responseline);
        return ENOTENMEM;
    }
    bzero(status,sizeof(struct status));
    if (break_status_line(responseline,status) == EBADSYNTAX) {
        free(responseline);
        free(status);
        return EBADSYNTAX;
    }
    free(responseline);

    /* fill in reply structure */
    reply->status.version = status->version;
    reply->status.code = status->code;
    reply->status.msg = status->msg;
    reply->headerno = 0;
    free(status);

    /* search for headers */
    ptrcrlfaux = index((char *) ptrcrlf +2,'\r');
    while (ptrcrlfaux != NULL) {
        /* extract header line */
        if (strlen(ptrcrlf +2) > strlen(ptrcrlfaux)) {
            if ((headerline = strndup((char *) ptrcrlf +2,
                                 strlen(ptrcrlf +2) - strlen(ptrcrlfaux))) == NULL) {
                free(reply->status.msg);
                return ENOTENMEM;
            }
            /* parse header line */
            if ((hdr = malloc(sizeof(struct header))) == NULL) {
                free(reply->status.msg);
                free(headerline);
                return ENOTENMEM;
            }
            bzero(hdr,sizeof(struct header));
            headerrslt = break_header(headerline,hdr,&reply->flags);
            if (headerrslt == EBADSYNTAX) {
                free(hdr->value);
                free(hdr);
                return EBADSYNTAX;
            }

            if (hdr->name == UNKNOWN)
                pos = add_header(reply,REPLY,hdr->name,(const u_char *) headerline);
            else
                pos = add_header(reply,REPLY,hdr->name,(const u_char *) hdr->value);
            free(headerline);
            free(hdr->value);
            free(hdr);

            /* search for new header */
            if (strncmp(ptrcrlfaux,"\r\n",2) != 0)
                return EBADSYNTAX;

            ptr = strstr((char *) ptrcrlfaux +2,"\r\n");
            ptrcrlf = ptrcrlfaux;
            ptrcrlfaux = ptr;
        } else
            /* length is equal, \r\n\r\n found? */
            ptrcrlfaux = NULL;
    }
    /* finished examining headers, body? */
    if (strncmp(ptrcrlf,"\r\n\r\n",4) != 0)	/* we know ptrcrlf is at least \r\n\n\0 */
        return EBADSYNTAX;

    ptrcrlf += 4;
    if (*ptrcrlf == '\0')
        reply->body = NULL;
    else
        if ((reply->body = strndup(ptrcrlf,strlen(ptrcrlf))) == NULL) {
            free(reply->status.msg);
            return ENOTENMEM;
        }

    return 0;
}

int break_status_line(const u_char *line, status_t status)
{
    int linelen, version, found;
    u_char *ptrb, *ptrmsg, *ptrcode, *ptrversion, *ptrcodetmp;

    assert(line != NULL);

    linelen = strlen(line);

    if (linelen == 0)
        return EBADSYNTAX;
    if (status == NULL)
        return ENALLOC;

    ptrb = index(line,' ');	/* search for a white */
    if (ptrb == NULL)
        return EBADSYNTAX;

    /* take version apart */
    if ((ptrversion = strndup(line,strlen(line) - strlen(ptrb))) == NULL)
        return ENOTENMEM;
    
    /* trim white spaces before code begins */
    while (*ptrb == ' ')
        if (strlen(ptrb) > 0)
            ptrb++;

    if (*(ptrcodetmp = ptrb) == '\0')    /* if empty, return */
        return EBADSYNTAX;

    ptrb = index(ptrcodetmp,' ');    /* search for a new white space */
    if (ptrb == NULL)
        return EBADSYNTAX;

    if ((ptrcode = strndup(ptrcodetmp,strlen(ptrcodetmp) - strlen(ptrb))) == NULL) {
        free(ptrversion);
        return ENOTENMEM;
    }

    /* trim whites */
    while (*ptrb == ' ')
        if (strlen(ptrb) > 0)
            ptrb++;

    /* check message */
    if (*(ptrmsg = ptrb) == '\0')    /* empty? */
        return EBADSYNTAX;

    if (strncasecmp(ptrversion,"HTTP/1.0",strlen(ptrversion)) == 0)
        version = HTTP10;
    else if (strncasecmp(ptrversion,"HTTP/1.1",strlen(ptrversion)) == 0)
        version = HTTP11;
    else
        version = UNKNOWN;
    free(ptrversion);

    found = 0;
    /* finally, check if code is ok */
    if (!found && strncasecmp(ptrcode,"100",strlen(ptrcode)) == 0) {
        status->code = CONTINUE;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"101",strlen(ptrcode)) == 0) {
        status->code = SWITCHPROTO;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"200",strlen(ptrcode)) == 0) {
        status->code = OK;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"201",strlen(ptrcode)) == 0) {
        status->code = CREATED;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"202",strlen(ptrcode)) == 0) {
        status->code = ACCEPTED;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"203",strlen(ptrcode)) == 0) {
        status->code = NONAUTHINFO;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"204",strlen(ptrcode)) == 0) {
        status->code = NOCONTENT;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"205",strlen(ptrcode)) == 0) {
        status->code = RSTCONTENT;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"206",strlen(ptrcode)) == 0) {
        status->code = PARCONTENT;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"300",strlen(ptrcode)) == 0) {
        status->code = MULCHOICES;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"301",strlen(ptrcode)) == 0) {
        status->code = MOVEDPERM;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"302",strlen(ptrcode)) == 0) {
        status->code = FOUND;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"303",strlen(ptrcode)) == 0) {
        status->code = SEEOTHER;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"304",strlen(ptrcode)) == 0) {
        status->code = NOTMODIFIED;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"305",strlen(ptrcode)) == 0) {
        status->code = USEPROXY;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"307",strlen(ptrcode)) == 0) {
        status->code = TEMPREDIR;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"400",strlen(ptrcode)) == 0) {
        status->code = BADREQUEST;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"401",strlen(ptrcode)) == 0) {
        status->code = UNAUTHORIZED;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"402",strlen(ptrcode)) == 0) {
        status->code = PAYREQUIRED;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"403",strlen(ptrcode)) == 0) {
        status->code = FORBIDDEN;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"404",strlen(ptrcode)) == 0) {
        status->code = NOTFOUND;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"405",strlen(ptrcode)) == 0) {
        status->code = METHODNALLOW;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"406",strlen(ptrcode)) == 0) {
        status->code = NOTACCEPT;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"407",strlen(ptrcode)) == 0) {
        status->code = PROXYAUTHREQ;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"408",strlen(ptrcode)) == 0) {
        status->code = REQTIMEOUT;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"409",strlen(ptrcode)) == 0) {
        status->code = CONFLICT;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"410",strlen(ptrcode)) == 0) {
        status->code = GONE;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"411",strlen(ptrcode)) == 0) {
        status->code = LENGTHREQ;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"412",strlen(ptrcode)) == 0) {
        status->code = PRECONDFAIL;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"413",strlen(ptrcode)) == 0) {
        status->code = REQENTLARGE;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"414",strlen(ptrcode)) == 0) {
        status->code = REQURILARGE;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"415",strlen(ptrcode)) == 0) {
        status->code = NMEDIATYPE;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"416",strlen(ptrcode)) == 0) {
        status->code = BADRANGE;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"417",strlen(ptrcode)) == 0) {
        status->code = EXPECTFAIL;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"500",strlen(ptrcode)) == 0) {
        status->code = INTERROR;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"501",strlen(ptrcode)) == 0) {
        status->code = NOTIMPLEM;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"502",strlen(ptrcode)) == 0) {
        status->code = BADGATEWAY;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"503",strlen(ptrcode)) == 0) {
        status->code = SERVICEUNA;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"504",strlen(ptrcode)) == 0) {
        status->code = GWTIMEOUT;
        found = 1;
    }
    if (!found && strncasecmp(ptrcode,"505",strlen(ptrcode)) == 0) {
        status->code = BADHTTPVER;
        found = 1;
    }

    if (!found)
        status->code = UNKNOWN;
    status->version = version;
    if ((status->msg = strndup(ptrmsg,strlen(ptrmsg))) == NULL) {
        free(ptrcode);
        return ENOTENMEM;
    }

    free(ptrcode);
    return 0;
}

/*
 * REPLY STRUCTURE MANAGEMENT
 */

reply_t init_reply(void)
{
    reply_t reply;
    int i;

    if ((reply = malloc(sizeof(struct reply))) == NULL)
        return NULL;
    bzero(reply,sizeof(struct reply));
    if (reply != NULL) {
        reply->status.version = -1;
        reply->status.code = -1;
        reply->status.msg = NULL;
        reply->flags = 0;
        reply->headerno = 0; /* zero headers by default */
        for (i = 0; i < MAXHEADERS; i++) {
            reply->header[i].name = EMPTY;
            reply->header[i].value = NULL;
        }
        reply->body = NULL;
    }

    return reply;
}

int free_reply(reply_t *reply)
{
    int i;
    reply_t r;

    assert(reply != NULL);
    r = *reply;
    assert(r != NULL);

    if (r->status.msg != NULL)
        free(r->status.msg);
    if (r->body != NULL)
        free(r->body);
    if (r->headerno > 0)
        for (i=0; i < r->headerno; i++)
            if (r->header[i].value != NULL)
                free(r->header[i].value);
    free(r);

    return 0;
}

u_char *sprintr(reply_t reply)
{
    u_char *buffer;
    int n, i;

    assert(reply != NULL);
    if ((buffer = malloc(MAXBUFSIZE*sizeof(u_char))) == NULL)
        return NULL;
    bzero(buffer,MAXBUFSIZE);

    // print request line
    if (reply->status.version == HTTP11) {
        n = http_11_status(buffer,reply->status.code,MAXBUFSIZE);
        for (i = 0; i < reply->headerno; i++)
            if ((reply->header[i].name != EMPTY) &&
                (reply->header[i].value != NULL))
                http_11_header(buffer,
                               reply->header[i].name,
                               MAXBUFSIZE - strlen(buffer),
                               reply->header[i].value);
    }
    if (reply->status.version == HTTP10) {
        n = http_10_status(buffer,reply->status.code,MAXBUFSIZE);
        for (i = 0; i < reply->headerno; i++) {
            if ((reply->header[i].name != EMPTY) &&
                (reply->header[i].value != NULL))
                http_10_header(buffer,
                               reply->header[i].name,
                               MAXBUFSIZE - strlen(buffer),
                               reply->header[i].value);
        }
    }

    n = crlf(buffer,MAXBUFSIZE - strlen(buffer));

    return buffer;
}

