/*
 * HTTP message analysis library
 * File name: message_analysis.h
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: May 19, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef __MESSAGE_ANALYSIS_H
#define __MESSAGE_ANALYSIS_H

#include "uri_analysis.h"

/* These declarations follow HTTP1.1 especs, see RFC2616 */
/* valid METHOD identifiers */
#define OPTIONS    0
#define GET        1
#define HEAD       2
#define POST       3
#define PUT        4
#define DELETE     5
#define TRACE      6
#define CONNECT    7
/* end of METHOD idenfitiers */

#define MAXMETHODS 8

/* valid HTTP VERSION identifiers */
#define HTTP10    0
#define HTTP11    1
/* end of HTTP VERSION identifiers */

/* status codes for HTTP 1.0 and HTTP/1.1 */
/* succesful */
#define CONTINUE        100
#define SWITCHPROTO     101
#define OK              200
#define CREATED         201
#define ACCEPTED        202
#define NONAUTHINFO     203
#define NOCONTENT       204
#define RSTCONTENT      205
#define PARCONTENT      206
/* redirection */
#define MULCHOICES      300
#define MOVEDPERM       301
#define FOUND           302
#define MOVEDTEMP       302
#define SEEOTHER        303
#define NOTMODIFIED     304
#define USEPROXY        305
#define TEMPREDIR       307
/* client error */
#define BADREQUEST      400
#define UNAUTHORIZED    401
#define PAYREQUIRED     402
#define FORBIDDEN       403
#define NOTFOUND        404
#define METHODNALLOW    405
#define NOTACCEPT       406
#define PROXYAUTHREQ    407
#define REQTIMEOUT      408
#define CONFLICT        409
#define GONE            410
#define LENGTHREQ       411
#define PRECONDFAIL     412
#define REQENTLARGE     413
#define REQURILARGE     414
#define NMEDIATYPE      415
#define BADRANGE        416
#define EXPECTFAIL      417
/* server error */
#define INTERROR        500
#define NOTIMPLEM       501
#define BADGATEWAY      502
#define SERVICEUNA      503
#define GWTIMEOUT       504
#define BADHTTPVER      505
/* no more status codes */

/* valid header FIELD name identifiers */
 /* general header fields */
#define CACHECONTROL    0
#define CONNECTION      1
#define DATE            2
#define PRAGMA          3
#define TRAILER         4
#define TRANSFERENC     5
#define UPGRADE         6
#define VIA             7
#define WARNING         8
 /* request header fields */
#define ACCEPT          9
#define ACCEPTCHARSET   10
#define ACCEPTENC       11
#define ACCEPTLANG      12
#define AUTHORIZATION   13
#define EXPECT          14
#define FROM            15
#define HOST            16
#define IFMATCH         17
#define IFMODIFSINCE    18
#define IFNONEMATCH     19
#define IFRANGE         20
#define IFUNMODSINCE    21
#define MAXFORWARDS     22
#define PROXYAUTHOR     23
#define RANGE           24
#define REFERER         25
#define TE              26
#define USERAGENT       27
 /* entity header fields */
#define ALLOW           28
#define CONTENTENC      29
#define CONTENTLANG     30
#define CONTENTLEN      31
#define CONTENTLOC      32
#define CONTENTMD5      33
#define CONTENTRANGE    34
#define CONTENTTYPE     35
#define EXPIRES         36
#define LASTMODIFIED    37
 /* HTTP/1.0 and HTTP/1.1 response header fields */
#define LOCATION        38
#define SERVER          39
#define WWWAUTH         40
 /* HTTP/1.1 only response header fields */
#define ACCEPTRANGES    41
#define AGE             42
#define ETAG            43
#define PROXYAUTHEN     44
#define RETRYAFTER      45
#define VARY            46
 /* strange header fields */
#define PROXYCONN       47
#define KEEPALIVE       48
/* end of FIELD identiers */

#define MAXHEADERS      49    /* max number of headers in a petition */
#define EBADSYNTAX      -2    /* for badly built requests */
#define ENOTENMEM       -3    /* if there is not enough memory */
#define EMPTY           -4    /* for empty header fields */

/* return value for unknown methods, http version or headers */
#define UNKNOWN         -1

#define PETITION	0
#define REPLY		1

/* general header related flags */
#define IFCLOSE         1    /* Connection: close */
#define IFKEEPALIVE     2    /* Connection: Keep-alive */
#define IFHOST          4    /* Host: ______ */
#define IFCONTLEN       8    /* Content-Length: _____ */
#define IFAUTH          16   /* Authorization: _____ */
#define IFRANGEH        32   /* Range: ____ */

/* macros for flag testing */
#define H_CLOSE(flags) ((flags & IFCLOSE) == IFCLOSE)
#define H_KEEPALIVE(flags) ((flags & IFKEEPALIVE) == IFKEEPALIVE)
#define H_HOST(flags) ((flags & IFHOST) == IFHOST)
#define H_CONTENTLEN(flags) ((flags & IFCONTLEN) == IFCONTLEN)
#define H_AUTHORIZATION(flags) ((flags & IFAUTH) == IFAUTH)
#define H_RANGEH(flags) ((flags & IFRANGE) == IFRANGE)

/* type definitions */
typedef struct request  *request_t;
typedef struct status   *status_t;
typedef struct header   *header_t;
typedef struct petition *petition_t;
typedef struct reply    *reply_t;

/* request line structure */
struct request
{
    int method;        /* method identifier */
    URI uri;           /* complete uri structure */
    int version;       /* http version identifier */
};

/* status line structure */
struct status
{
    int    version;    /* http version identifier */
    int    code;       /* status code */
    u_char *msg;       /* message explaining code */
};

/* header structure */
struct header
{
    int    name;    /* header field identifier */
    u_char *value;  /* header field value */
};

/* petition structure */
struct petition
{
    struct request request;
    int            headerno;        /* how much headers available in header[] */
    struct header  header[MAXHEADERS];
    int            flags;           /* flags for common headers */
    u_char         *body;           /* message body (if any) */
};

/* reply structure */
struct reply
{
    struct status  status;
    int	           headerno;    /* how much headers available in header[] */
    struct header  header[MAXHEADERS];
    int            flags;       /* flags for common headers */
    u_char         *body;       /* temp FILENAME where body is stored (if any) */
};

/*
 * PETITION ANALYSIS FUNCS
 */

/*
 * Takes a string with a request and breaks it into a petition
 * structure.
 * Returns:
 *   EBADSYNTAX if request was incorrectly built.
 *   ENALLOC if petition was not previously allocated.
 *   -1 if any other error occurs (see errno).
 */
int analyze_petition(const u_char *request, petition_t petition);

/*
 * Breaks a request line into its fields, method, uri and http version.
 * Returns:
 *   EBADSYNTAX if line was incorrectly built (that is, no valid method
 *    or http version, invalid uri, or malformed line).
 *   ENALLOC if request was not previously allocated.
 *   -1 if any other error occurs (see errno).
 */
int break_request_line(const u_char *line, request_t request);

/*
 * Checks if a petition is well built regarding host specification.
 * If host is specified in uri and host header, they must be the same.
 * If host header includes port, check it out.
 * If port is ok, then check_host() assures the it corresponds
 * to the port specified by petition.
 * Returns:
 *   1 if a host is specified and is coherent,
 *   0 elsewhere.
 */
int check_host(const petition_t petition, int port);

/*
 * Returns a pointer to the actual domain which petition is refering to,
 * if there is one. This function assumes petition is coherent.
 * default_domain is returned if no host is specified.
 * This function will allocate memory which has to be freed.
 */
u_char *get_host(const petition_t petition);

/*
 * PETITION STRUCTURE MANAGEMENT
 */

/*
 * Initializes a petition structure. Allocates memory, and returns
 * a pointer to it.
 * Returns:
 *   NULL if there was an error.
 */
petition_t init_petition(void);

/*
 * Frees memory asociated with a petition struct.
 */
int free_petition(petition_t *petition);

/*
 * Prints a petition structure to a string, returning it.
 * The resulting string will be a valid HTTP petition.
 */
u_char *sprintp(petition_t petition);

/*
 * HEADERS MANAGEMENT
 */

/*
 * Breaks a header line into its name and value fields.
 * Returns:
 *   EBADSYNTAX if line was incorrectly built (that is, no valid header
 *    field or just it was incomplete or malformed).
 *   ENALLOC if header was not previously allocated.
 *   -1 if any other error occurs (see errno).
 */
int break_header(const u_char *line, header_t header, int *flag);

/*
 * Adds a determinate header to the header list of a petition or
 * reply. The kind of structure to use will be determined by type,
 * that can be PETITION or REPLY.
 * Returns:
 *   0 if success.
 *   -1 on error.
 */
int add_header(void *message,
               const int type,
	       const int name,
	       const u_char *value);

/*
 * Searchs for a header on a header list of a petition or reply.
 * The kind of structure to use will be determined by type, that
 * can be PETITION or REPLY.
 * Returns:
 *   position where header was found.
 *   message->headerno if not found.
 *   -1 on error.
 */
int search_header(void *message, const int type, const int name);

/*
 * Removes a determinate header from the header list of a petition or
 * reply. The kind of structure to use will be determined by type,
 * that can be PETITION or REPLY.
 * Returns:
 *   0 if success.
 *   -1 on error.
 */
int remove_header(void *message, const int type, const int name);

/*
 * REPLY ANALYSIS FUNCS
 */

/*
 * Takes a string with a response and breaks it into a reply
 * structure.
 * Returns:
 *   EBADSYNTAX if request was incorrectly built.
 *   ENALLOC if reply was not previously allocated.
 *   -1 if any other error occurs (see errno).
 */
int analyze_reply(const u_char *response, reply_t reply);

/*
 * Breaks a status line into its fields, version, code, and message fields.
 * Returns:
 *   EBADSYNTAX if line was incorrectly built (that is, no valid
 *    or http version or code).
 *   ENALLOC if status was not previously allocated.
 *   -1 if any other error occurs (see errno).
 */
int break_status_line(const u_char *line, status_t status);

/*
 * REPLY STRUCTURE MANAGEMENT
 */

/*
 * Initializes a reply structure. Allocates memory, and returns
 * a pointer to it.
 * Returns:
 *   NULL if there was an error.
 */
reply_t init_reply(void);

/*
 * Frees memory asociated with a reply struct.
 */
int free_reply(reply_t *reply);

/*
 * Prints a reply structure to a string, returning it.
 * The resulting string will be a valid HTTP reply.
 */
u_char *sprintr(reply_t reply);

#endif    /* message_analysis.h */

