/*
 * HTTP Proxy server implementation
 * File name: proxy.c
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: June 22, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "proxy.h"
#include "message_analysis.h"
#include "http_10.h"
#include "http_11.h"
#include "statistics.h"

int is_proxy(const petition_t petition)
{
    if (petition->request.uri.host != NULL) {
        if (is_localdomain(petition->request.uri.host)) {
            if (petition->request.uri.port_present) {
                if (petition->request.uri.port == s_port)
                    return 0;
            } else
                return 0;
        }
    } else
        return 0;	// host error, nothing to do here!

    return 1;	// all ok
}

int attend_remote_petition(const int fd, petition_t petition, const int headers)
{
    struct sockaddr_in server;
    struct hostent *he;
    struct timeval *tv;
    int httpver, sock, n, pos;
    long int contentlen;
    u_char *request = NULL, *response = NULL;
    reply_t reply;

    assert(petition != NULL);
    assert(petition->request.uri.host != NULL);
    httpver = petition->request.version;

    // assure host is ok
    if (!check_host(petition,0)) {
        if (httpver == HTTP10) {
            write_log(fd,petition,BADREQUEST,"default");
            http_10_errmsg(fd,BADREQUEST,NULL,headers,petition->request.method);
        } else {
            write_log(fd,petition,BADREQUEST,"default");
            http_11_errmsg(fd,BADREQUEST,NULL,headers,petition->request.method);
        }
        return 0;
    }

    // if petition is ok, then increase remote petitions counter
    increase_remote();

    // resolve hostname
    if ((he = gethostbyname(petition->request.uri.host)) == NULL) {
        if (httpver == HTTP10) {
            write_log(fd,petition,NOTFOUND,"default");
            http_10_errmsg(fd,NOTFOUND,NULL,headers,petition->request.method);
        } else {
            write_log(fd,petition,NOTFOUND,"default");
            http_11_errmsg(fd,NOTFOUND,NULL,headers,petition->request.method);
        }
        return 0;
    }

    // create socket
    if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
        fprintf(stderr,"Failed to create socket\n");
        if (httpver == HTTP10) {
            write_log(fd,petition,INTERROR,"default");
            http_10_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        } else {
            write_log(fd,petition,INTERROR,"default");
            http_11_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        }
        return 0;
    }

    if ((tv = malloc(sizeof(struct timeval))) == NULL)
        return -1;
    bzero(tv,sizeof(struct timeval));
    tv->tv_sec = get_timeout();
    tv->tv_usec = 0;
    setsockopt(sock,SOL_SOCKET,SO_RCVTIMEO,tv,sizeof(struct timeval));
    setsockopt(sock,SOL_SOCKET,SO_SNDTIMEO,tv,sizeof(struct timeval));
    free(tv);

    /* build the server structure */
    bzero(&server,sizeof(server));
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = *((unsigned long *) he->h_addr_list[0]);
    server.sin_port = htons(petition->request.uri.port);

    /* try server */
    if (connect(sock,(struct sockaddr *) &server, sizeof(server)) < 0) {
        if (httpver == HTTP10) {
            write_log(fd,petition,BADGATEWAY,"default");
            http_10_errmsg(fd,BADGATEWAY,NULL,headers,petition->request.method);
        }
        if (httpver == HTTP11) {
            write_log(fd,petition,GWTIMEOUT,"default");
            http_11_errmsg(fd,GWTIMEOUT,NULL,headers,petition->request.method);
        }
        return 0;
    }

    /* remove connection related headers and flags */
    remove_header(petition,PETITION,CONNECTION);
    remove_header(petition,PETITION,PROXYCONN);
    remove_header(petition,PETITION,KEEPALIVE);
    remove_header(petition,PETITION,TE);
    petition->flags &= !IFCLOSE && !IFKEEPALIVE;
    add_header(petition,PETITION,CONNECTION,"close");

    /* build request string */
    if ((request = sprintp(petition)) == NULL) {
        if (httpver == HTTP10) {
            write_log(fd,petition,INTERROR,"default");
            http_10_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        }
        if (httpver == HTTP11) {
            write_log(fd,petition,INTERROR,"default");
            http_11_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        }
        return 0;
    }

    /* send request */
    n = writen(sock,request,strlen(request));
    free(request);

    // did the request reach destination?
    if (n == -1) {
        if (httpver == HTTP10) {
            write_log(fd,petition,BADGATEWAY,"default");
            http_10_errmsg(fd,BADGATEWAY,NULL,headers,petition->request.method);
        }
        if (httpver == HTTP11) {
            write_log(fd,petition,GWTIMEOUT,"default");
            http_11_errmsg(fd,GWTIMEOUT,NULL,headers,petition->request.method);
        }
        return 0;
    }

    /* receive response */
    if ((response = malloc(MAXBUFSIZE*sizeof(u_char))) == NULL)
        return -1;

    bzero(response,MAXBUFSIZE);
    if (readn(sock,response,MAXBUFSIZE) == -1) {
        if (httpver == HTTP10) {
            write_log(fd,petition,BADGATEWAY,"default");
            http_10_errmsg(fd,BADGATEWAY,NULL,headers,petition->request.method);
        }
        if (httpver == HTTP11) {
            write_log(fd,petition,GWTIMEOUT,"default");
            http_11_errmsg(fd,GWTIMEOUT,NULL,headers,petition->request.method);
        }
        return 0;
    }

    /* assure reply is ok */
    if ((reply = init_reply()) == NULL) {
        fprintf(stderr,"attend_remote_petition(): %s\n",strerror(errno));
        if (httpver == HTTP10) {
            write_log(fd,petition,INTERROR,"default");
            http_10_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        }
        if (httpver == HTTP11) {
            write_log(fd,petition,INTERROR,"default");
            http_11_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        }
        free(response);
        return 0;
    }
    if (analyze_reply(response,reply) == EBADSYNTAX) {
        if (httpver == HTTP10) {
            write_log(fd,petition,BADGATEWAY,"default");
            http_10_errmsg(fd,BADGATEWAY,NULL,headers,petition->request.method);
        }
        if (httpver == HTTP11) {
            write_log(fd,petition,BADGATEWAY,"default");
            http_11_errmsg(fd,BADGATEWAY,NULL,headers,petition->request.method);
        }
        free(response);
        return 0;
    }

    // check length of body
    contentlen = MAXBUFSIZE;
    pos = search_header(reply,REPLY,CONTENTLEN);
    if ((pos != -1) && (pos != reply->headerno))
        if (reply->header[pos].value != NULL)
            contentlen = strtol(strndup(reply->header[pos].value,strlen(reply->header[pos].value)),NULL,10);
    contentlen = (contentlen < MAXBUFSIZE) ? contentlen : MAXBUFSIZE;

    remove_header(reply,REPLY,CONNECTION);

    /* add connection headers if needed */
    if (H_CLOSE(headers))
        add_header(reply,REPLY,CONNECTION,"close");

    /* build response to client */
    free(response);
    if ((response = sprintr(reply)) == NULL) {
        if (httpver == HTTP10) {
            write_log(fd,petition,INTERROR,"default");
            http_10_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        }
        if (httpver == HTTP11) {
            write_log(fd,petition,INTERROR,"default");
            http_11_errmsg(fd,INTERROR,NULL,headers,petition->request.method);
        }
        free_reply(&reply);
        return 0;
    }
    free_reply(&reply);

    /* send response http header */
    if (writen(fd,response,strlen(response)) == -1) {
        free(response);
        close(sock);
        return -1;
    }
    

    /* read body and send to client */
    bzero(response,MAXBUFSIZE);
    while ((n = read(sock,response,contentlen)) != 0) {
        if ((n == -1) && (errno != EINTR)) {
            if (httpver == HTTP10) {
                write_log(fd,petition,BADGATEWAY,"default");
                http_10_errmsg(fd,BADGATEWAY,NULL,headers,petition->request.method);
            }
            if (httpver == HTTP11) {
                write_log(fd,petition,GWTIMEOUT,"default");
                http_11_errmsg(fd,GWTIMEOUT,NULL,headers,petition->request.method);
            }
            break;
        }
        if (writen(fd,response,n) == -1) {
            free(response);
            close(sock);
            return -1;
        }
        bzero(response,MAXBUFSIZE);
    }

    close(sock);
    free(response);

    return 0;
}

