/*
 * HTTP server statistics implementation
 * File name: statistics.c
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: May 19, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "statistics.h"
#include "http_10.h"
#include "http_11.h"
#include "http_config.h"

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t log  = PTHREAD_MUTEX_INITIALIZER;

/*
 * Takes account of how many local, remote or bad syntax
 * petitions arrive.
 */
static struct counters counters;

/*
 * The file where to log.
 */
static int logfd;

void init_stats(void)
{
    counters.local_petitions = 0;
    counters.remote_petitions = 0;
    counters.badsyntax_petitions = 0;
}

void increase_local(void)
{
    pthread_mutex_lock(&lock);
    counters.local_petitions++;
    pthread_mutex_unlock(&lock);
}

void increase_remote(void)
{
    pthread_mutex_lock(&lock);
    counters.remote_petitions++;
    pthread_mutex_unlock(&lock);
}

void increase_badsyntax(void)
{
    pthread_mutex_lock(&lock);
    counters.badsyntax_petitions++;
    pthread_mutex_unlock(&lock);
}

counters_t return_counters(void)
{
    counters_t c;

    if ((c = malloc(sizeof(struct counters))) != NULL) {
        bzero(c,sizeof(struct counters));

        pthread_mutex_lock(&lock);
        c->local_petitions = counters.local_petitions;
        c->remote_petitions = counters.remote_petitions;
        c->badsyntax_petitions = counters.badsyntax_petitions;
        pthread_mutex_unlock(&lock);
    }
    return c;
}

void reset_stats(void)
{
    pthread_mutex_lock(&lock);
    init_stats();
    pthread_mutex_unlock(&lock);
}

void init_logs(void)
{
    char buf[MAXBUFSIZE];

    logfd = open(get_logfile(), O_CREAT | O_APPEND | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP);
    if (logfd == -1) {
        fprintf(stderr,"init_logs(): %s %s\n",strerror(errno),get_logfile());
        _exit(EXIT_FAILURE);
    }
    snprintf(buf,MAXBUFSIZE,"%s started at %s (%s:%ld) with UID %d\n",
            SERVERNAME,
            getenv("HOSTNAME"),
            inet_ntoa(s_addr),
            s_port,
            getuid());
    writen(logfd,buf,strlen(buf));
}

int write_log(const int fd, const petition_t petition, const int code, const char *message)
{
    char buf[MAXBUFSIZE], line[MAXBUFSIZE], *t, uri[MAXBUFSIZE];
    struct sockaddr_in *sa;
    struct timeval *tv;
    struct timezone *tz;
    int n;

    bzero(buf,MAXBUFSIZE);
    bzero(line,MAXBUFSIZE);
    bzero(uri,MAXBUFSIZE);

    // write uri to buffer
    if (petition->request.uri.port_present)
        snprintf(uri,MAXBUFSIZE,"%s%s:%d%s",
                 petition->request.uri.protocol,
                 petition->request.uri.host,
                 petition->request.uri.port,
                 petition->request.uri.path);
    else
        snprintf(uri,MAXBUFSIZE,"%s%s%s",
                 (petition->request.uri.host != NULL) ? petition->request.uri.protocol : (u_char *) "",
                 (petition->request.uri.host != NULL) ? petition->request.uri.host : (u_char *) "",
                 petition->request.uri.path);

    // write request line to buffer
    switch (petition->request.version) {
    case HTTP10:
        if (http_10_request(buf,petition->request.method,uri,MAXBUFSIZE) == ENOTENMEM)
            return ENOTENMEM;
        buf[strlen(buf) - 2] = '\0';    // trim trailing \r\n
        break;
    case HTTP11:
        if (http_11_request(buf,petition->request.method,uri,MAXBUFSIZE) == ENOTENMEM)
            return ENOTENMEM;
        buf[strlen(buf) - 2] = '\0';    // trim trailing \r\n
        break;
    default:
        switch (petition->request.method) {
        case GET:
            snprintf(buf,MAXBUFSIZE,"GET %s UNKNOWN",uri);
            break;
        case HEAD:
            snprintf(buf,MAXBUFSIZE,"HEAD %s UNKNOWN",uri);
            break;
        case POST:
            snprintf(buf,MAXBUFSIZE,"POST %s UNKNOWN",uri);
            break;
        default:
            snprintf(buf,MAXBUFSIZE,"UNKNOWN %s UNWNOWN",uri);
        }
    }

    // write remote peer address to a buffer
    n = sizeof(struct sockaddr_in);
    if ((sa = malloc(n)) == NULL)
        return -1;
    // this is a dirty hack, getpeername expects struct sockaddr, not struct sockaddr_in
    getpeername(fd,(struct sockaddr *) sa,&n);

    // write actual time and trim \n
    if ((t = malloc(26*sizeof(char))) == NULL) {
        free(sa);
        return -1;
    }
    if ((tv = malloc(sizeof(struct timeval))) == NULL) {
        free(sa);
        free(t);
        return -1;
    }
    if ((tz = malloc(sizeof(struct timezone))) == NULL) {
        free(sa);
        free(t);
        free(tv);
        return -1;
    }
    gettimeofday(tv,tz);
    t = ctime_r(&tv->tv_sec,t);
    t[strlen(t) - 1] = '\0';
    
    snprintf(line,MAXBUFSIZE,"%s - [%s] \"%s\" %d \"%s\"\n",
             inet_ntoa(sa->sin_addr),t,buf,code,message);
    free(sa);
    free(tz);
    free(tv);
    free(t);
    
    pthread_mutex_lock(&log);
    n = write(logfd,line,strlen(line));
    pthread_mutex_unlock(&log);

    return n;
}
