/*
 * URI library implementation
 * File name: uri_analysis.c
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: April 25, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "utils.h"
#include "uri_analysis.h"

int break_uri (const u_char *uri, uri_t elements)
{
    int urilen;
    u_char http[] = "http://";
    u_char *ptrhost, *ptrport, *ptrpath, *port;

    urilen = strlen(uri);

    if (urilen == 0)
        return EBADURI;

    if (elements == NULL)
        return ENALLOC;

    /* uri has one or more characters */

    /* relative uri! */
    if (*uri == '/') {
        if (elements != NULL) {
            if ((elements->protocol = strndup(http,7)) == NULL)
                return ENOTMEM;

            elements->host = NULL;
            elements->port = 0;	/* use -1 to remark it's not absolute uri */
            if ((elements->path = strndup(uri,urilen)) == NULL) {
                free(elements->protocol);
                return ENOTMEM;
            }
            if (test_str(uri,urilen,0))
                return ENOTSEC;
        } else
            return (-1);
    /* absolute uri! */
    } else {
        if (urilen < 8)		/* http://_ is minimum */
            return EBADURI;

        if (strncmp(uri,http,7) != 0)
            return EBADURI;	/* http:// must be preceding uri */

        ptrhost = (u_char *) uri + 7*sizeof(u_char);	/* points to a host field on an uri */
        ptrport = index(ptrhost,':');			/* search for port */
        ptrpath = index(ptrhost,'/');			/* search for path */
        if (ptrport == NULL)
            /* no port nor path, using 80 and / respectively */
            if (ptrpath == NULL) {
                if ((elements->protocol = strndup(http,7)) == NULL)
                    return ENOTMEM;

                if ((elements->host = strndup(ptrhost,urilen - 7)) == NULL) {
                    free(elements->protocol);
                    return ENOTMEM;
                }
                elements->port = 80;
                if ((elements->path = malloc(2*sizeof(u_char))) == NULL) {
                    free(elements->protocol);
                    free(elements->host);
                    return ENOTMEM;
                }
                bzero(elements->path,2*sizeof(u_char));
                snprintf(elements->path,2,"/");

            /* no port but path present */
            } else {
                if ((elements->protocol = strndup(http,7)) == NULL)
                    return ENOTMEM;

                if ((elements->host = strndup(ptrhost,strlen(ptrhost) - strlen(ptrpath))) == NULL) {
                    free(elements->protocol);
                    return ENOTMEM;
                }
                elements->port = 80;
                if ((elements->path = strndup(ptrpath,strlen(ptrpath))) == NULL) {
                    free(elements->protocol);
                    free(elements->host);
                    return ENOTMEM;
                }
                if (test_str(ptrpath,strlen(ptrpath),0))
                    return ENOTSEC;
            }
        else
            /* port found but no path */
            if (ptrpath == NULL) {
                if ((elements->protocol = strndup(http,7)) == NULL)
                    return ENOTMEM;

                if ((elements->host = strndup(ptrhost,strlen(ptrhost) - strlen(ptrport))) == NULL) {
                    free(elements->protocol);
                    return ENOTMEM;
                }
                if ((elements->path = malloc(2*sizeof(u_char))) == NULL) {
                    free(elements->protocol);
                    free(elements->host);
                    return ENOTMEM;
                }
                bzero(elements->path,2*sizeof(u_char));
                snprintf(elements->path,2,"/");
                elements->port = strtol(ptrport + 1*sizeof(u_char),NULL,10);
                elements->port_present = 1;

            /* port and path found */
            } else {
                if ((elements->protocol = strndup(http,strlen(http))) == NULL)
                    return ENOTMEM;

                if ((elements->host = strndup(ptrhost,strlen(ptrhost) - strlen(ptrport))) == NULL) {
                    free(elements->protocol);
                    return ENOTMEM;
                }
                if ((elements->path = strndup(ptrpath,strlen(ptrpath))) == NULL) {
                    free(elements->protocol);
                    free(elements->host);
                    return ENOTMEM;
                }
                if ((port = strndup(ptrport + 1*sizeof(u_char),strlen(ptrport + 1*sizeof(u_char)) - strlen(ptrpath))) == NULL) {
                    free(elements->protocol);
                    free(elements->host);
                    free(elements->path);
                    return ENOTMEM;
                }
                elements->port = strtol(port,NULL,10);
                elements->port_present = 1;
                free(port);
                if (test_str(ptrpath,strlen(ptrpath),0))
                    return ENOTSEC;
            }
    }
    return 0;
}

uri_t init_uri(void)
{
    uri_t uri;

    if ((uri = malloc(sizeof(struct uri))) != NULL)
        bzero(uri,sizeof(struct uri));

    return uri;
}

int free_uri(uri_t *uri)
{
    uri_t u;

    assert(uri != NULL);
    u = *uri;
    assert(u != NULL);

    if (u->protocol != NULL)
        free(u->protocol);
    if (u->host != NULL)
        free(u->host);
    if (u->path != NULL)
        free(u->path);
    free(u);

    return 0;
}
