/*
 * General purpouse utilities implementation
 * File name: utils.c
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: May 2, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "utils.h"

ssize_t readn (int fd, void *vptr, size_t n)
// read n bytes from file descriptor
{
    size_t  nleft;
    ssize_t nread;
    u_char  *ptr;
    int     trim = 1;

    ptr = (u_char *) vptr;
    nleft = n;

    while (nleft > 0) {
        if ((nread = read(fd, ptr, (nleft > 0) ? 1 : 0)) < 0) {
            if (errno == EINTR) {
                nread = 0;      // retry
                continue;
            } else
                return (-1);
        } else if (nread == 0)
                break;              // got EOF

        /* trim \r and \n characters at the begining of the petition */
        if (trim) {
            if (*ptr != '\r' && *ptr != '\n') {
                nleft -= nread;
                ptr += nread;
                trim = 0;
            }
        } else {
            nleft -= nread;
            ptr += nread;
        }

        if ((n - nleft) >= 4)
            if (strncmp((u_char *) ptr - 4,"\r\n\r\n",4) == 0)
                return (n - nleft);
        if ((n - nleft) >= 2)
            if (strncmp((u_char *) ptr - 2,"\n\n",2) == 0)
                return (n - nleft);
    }
    return (n - nleft);         // return read
}

ssize_t writen (int fd, const void *vptr, size_t n)
// Write n bytes to file descriptor
{
    size_t nleft;
    ssize_t nwritten;
    const char *ptr;

    ptr = (const char *) vptr;
    nleft = n;
    while (nleft > 0) {
        if ((nwritten = write(fd, ptr, nleft)) <= 0) {
            if ((nwritten == -1) && (errno == EINTR)) {
                nwritten = 0;   // retry
                continue;
            } else
                return (-1);    // error!
        }
        nleft -= nwritten;
        ptr += nwritten;
    }
    return (n);
}

ssize_t crlf(u_char *str, size_t n)
{
    char *tmp;
    size_t bytes = 0;
    
    if ((tmp = malloc(3*sizeof(u_char))) != NULL) {
        bzero(tmp,3*sizeof(u_char));
        bytes = snprintf(tmp,3,"\r\n");
        strncat(str,tmp,(n < 3) ? n : 3);
        free(tmp);
    }

    return bytes;
}

u_char *erase_slash(u_char *str)
{
    int len;

    if (str == NULL)
        return NULL;;

    len = strlen(str);
    if (*(str + (len - 1)*sizeof(u_char)) == '/')
        *(str + (len - 1)*sizeof(u_char)) = '\0';

    return str;
}

off_t get_bodylen(int file)
{
    int i, found;
    ssize_t nread;
    off_t fsize, n;
    u_char *ptr;
    u_char buf[MAXBUFSIZE];
    struct stat *filestats;

    /* initializations */
    n = 0;
    found = 0;
    if ((filestats = malloc(sizeof(struct stat))) == NULL)
        return n;
    assert(fstat(file,filestats) != -1);
    fsize = filestats->st_size;
    free(filestats);

    /* assure we are reading from the begining of the file */
    lseek(file, 0, SEEK_SET);
   
    /* read the contents of the file, searching two consecutive line feeds */
    while ((nread = read(file, buf, MAXBUFSIZE)) != 0) {
        if ((nread == -1) && (errno = EINTR))
            continue;
        ptr = buf;
        for (i = 0; i < nread; i++) {
            if (found)
                n++;
            if (i >= 1)
                /* if cgi printed headers, it should print at least one \n\n */
                if ((!found) && (strncmp(ptr - 1*sizeof(u_char),"\n\n",2) == 0))
                    found = 1;
            ptr++;
        }
    }

    return (n == 0) ? fsize : n;
}

char *dir2html(DIR *dir, u_char *filename, u_char *wd)
{
    char *buf;
    struct dirent *entry;
    struct stat *stats;
    char *lastmod;
    char path[MAXPATHNAME];
    int i, n;

    assert(dir != NULL);
    assert(filename != NULL);

    if ((buf = malloc(MAXBUFSIZE*sizeof(char))) == NULL)
        return NULL;

    if ((stats = malloc(sizeof(struct stat))) == NULL) {
        free(buf);
        return NULL;
    }

    // print HTML header
    snprintf(buf,MAXBUFSIZE,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n\
<html>\n <head>\n  <title>Index of %s</title>\n </head>\n <body>\n  <h1>Index of %s</h1>\n  <pre>\n Name\t\t\t\
Last Modified\t\t\tSize\n<hr/>",
             filename,filename);

    entry = readdir(dir);
    while (entry != NULL) {
        i = 0;
        if (strcmp(entry->d_name,".") != 0) {
            bzero(stats,sizeof(struct stat));
            bzero(path,MAXPATHNAME);
            // create relative pathname to dir entry
            snprintf(path,MAXPATHNAME,"%s/%s",wd,entry->d_name);
            
            // check stats for entry
            stat(path,stats);
            if ((lastmod = malloc(26*sizeof(char))) == NULL) {
                free(buf);
                free(stats);
                return NULL;
            }
            lastmod = ctime_r(&stats->st_mtime,lastmod);
            // erase trailing \n
            if (lastmod != NULL)
                *(lastmod + (strlen(lastmod) - 1)*sizeof(char)) = '\0';

            // print HTML template
            snprintf(buf + strlen(buf),MAXBUFSIZE - strlen(buf)," <a href=\"%s/%s\">",
                     erase_slash(filename),entry->d_name);
            if (strlen(entry->d_name) < 22)
                snprintf(buf + strlen(buf),(strlen(entry->d_name) +1 < (MAXBUFSIZE - strlen(buf))) ? 
                         strlen(entry->d_name) +1 : 
                         (MAXBUFSIZE - strlen(buf)),
                         "%s%n",entry->d_name,&n);
            else {
                snprintf(buf + strlen(buf),(20 < (MAXBUFSIZE - strlen(buf)))
                         ? 20 : (MAXBUFSIZE - strlen(buf)),"%s%n",entry->d_name,&n);
                snprintf(buf + strlen(buf),MAXBUFSIZE - strlen(buf),"...");
            }

            snprintf(buf + strlen(buf),MAXBUFSIZE - strlen(buf),"</a> ");
            for (i = 0; i < (22 - n); i++)
                snprintf(buf + strlen(buf),MAXBUFSIZE - strlen(buf)," ");

            snprintf(buf + strlen(buf),MAXBUFSIZE - strlen(buf),"%s\t%d\n",lastmod,(int) stats->st_size);

            free(lastmod);
        }
        entry = readdir(dir);
    }

    // print HTML footer
    snprintf(buf + strlen(buf),MAXBUFSIZE - strlen(buf),"  <hr/></pre>\n\
  <address>%s</address>\n </body>\n</html>",SERVERNAME);

    free(stats);
    free(entry);
    return buf;
}

int test_str(const u_char *str, int len, int tests)
{
    int rslt = SEC_OK, i;
    u_char *ptr;

    ptr = (u_char *) str;
    
    /* search for a "previous directory" string */
    if (strstr(str,"/../") != NULL)
        rslt |= SEC_BACKWARDS;
    
    /* search for '\0' characters */
    if (S_IFNULLBYTES(tests) || tests == 0)
        for (i = 0; i < len; i++)
            if (str[i] == '\0') {
                rslt |= SEC_NULLBYTES;
                break;
            }

    return rslt;
}

