/*
 * General purpouse utilities & definitions library
 * File name: utils.h
 * Author: Jaime Prez Crespo <japecre@lsub.org>
 * Last modification: May 20, 2004
 *
 * Copyright (c) 2004, Jaime Prez Crespo
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list
 *   of conditions and the following disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 * - Neither the name of the Rey Juan Carlos University nor the names of its contributors
 *   may be used to endorse or promote products derived from this software without specific
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef __UTILS_H
#define __UTILS_H

#include "defs.h"
#include "http_config.h"

/*
 * String management utilities
 */

/*
 * Reads max n bytes from fd and put them at vptr.
 * Stops if two \n found.
 * Returns bytes read, -1 if error.
 */
ssize_t readn(int fd, void *vptr, size_t n);

/*
 * Writes n bytes pointed by vptr into fd descriptor.
 * Returns bytes written, -1 if error.
 */
ssize_t writen(int fd, const void *vptr, size_t n);

/*
 * Concatenates max n bytes from a CRLF to a string. Returns bytes written.
 */
ssize_t crlf(u_char *str, size_t n);

/*
 * Erases trailing '/' character from a string.
 * Returns 0 upon success, -1 upon failure.
 */
u_char *erase_slash(u_char *str);

/*
 * File utilities
 */

/*
 * Searchs for \n\n and returns file's length from there.
 * Returns
 *   file whole length if no \on\n sequence found.
 */
off_t get_bodylen(int file);

/*
 * Returns a buffer with HTML code describing the contents of
 * a directory.
 * filename is the server path to the directory to list.
 * wd is the server internal path to the directory.
 */
char *dir2html(DIR *dir, u_char *filename, u_char *wd);

/*
 * Security utilities
 */

/* Some security related constants */
#define SEC_OK           0    // no malicious strings
#define SEC_BACKWARDS    1    // "/../" like string found
#define SEC_NULLBYTES    2    // null character (\0) found

/* macros for security checks */
#define S_IFBACKWARDS(flags)    ((flags & SEC_BACKWARDS) == SEC_BACKWARDS)
#define S_IFNULLBYTES(flags)    ((flags & SEC_NULLBYTES) == SEC_NULLBYTES)

/*
 * Performs some sanity checks over a string with length len.
 * The tests to do can be defined by XOR'ing the desired tests flags when
 * calling this function. All will be performed if tests is zero valued.
 * Returns:
 *   0 if OK.
 *   !0 if malicious strings found, being XOR'ed.
 */
int test_str(const u_char *str, int len, int tests);

#endif    /* utils.h */

